#ifdef VMS
#pragma nostandard
#endif
#include <dce/idlddefs.h>
#include <lw/types.h>
#include <lw/security-types.h>
#include <lw/rpc/dssetup.h>
#include <lw/rpc/wkssvc.h>
#include <lw/rpc/lsa.h>

#include "srvsvc.h"
#include "dssetup.h"
#include "wkssvc.h"
#include "lsa.h"

#include "rpcservice.h"

#include "hvrpc.h"
#include "hvrpcint.h"

#define NDR_DREP_INT_REP(drep)   ((drep)[0] >> 4)
#define NDR_DREP_CHAR_REP(drep)  ((drep)[0] & 0xf)
#define NDR_DREP_FLOAT_REP(drep) ((drep)[1])

#define DCERPC_PKT_RESPONSE 2
#define DCERPC_PFC_FLAG_FIRST 1
#define DCERPC_PFC_FLAG_LAST  2


#define SWAP16(val) hvswap16((unsigned8 *)&(val))
#define SWAP32(val) hvswap32((unsigned8 *)&(val))
#define SWAP(x, y) {unsigned8 tmp = x; x = (y); y = tmp;}

void hvswap16(unsigned8 *val16_p);
void hvswap32(unsigned8 *val32_p);
int id_compare(const char *id1, const char *id2);

void *hvbuf_malloc(int64_t bytes);

int hvrpc_init() {
	return 0;
}

/*
 * hvrpc_lookup() is called to retreive the RPC endpoint.
 * Each endpoint can implement multiple interfaces.
 * In our implementation each interface is independant and there is
 * no two-level hierarchy of endpoint -> interface.
 * Hence this routine is free to return any non-zero value for each supported endpoint.
 */
hvrpc_iface_handle_t hvrpc_lookup(const char *service) {
	if (!strcasecmp(service, "SRVSVC")) {
		return srvsvc_v3_0_s_ifspec;
	} else if (!strcasecmp(service, "DSSETUP")) {
		return dssetup_v0_0_s_ifspec;
	} else if (!strcasecmp(service, "LSARPC")) {
		return dssetup_v0_0_s_ifspec;
	} else if (!strcasecmp(service, "WKSSVC")) {
		return wkssvc_v1_0_s_ifspec;
	}

	printf("INFO: Unknown service %s\n", service);
	return NULL;
}

/*
 * hvrpc_lookup_by_id() is called o retreive an interface for a given endpoint.
 * Since we dont have the endpoint -> interface hierarchy, we simply ignore the
 * endpoint argument and retreive each supported interface.
 */
hvrpc_iface_handle_t hvrpc_lookup_by_id(const char *endpoint, const char *p_interface_id) {
	if (!id_compare(p_interface_id, (char *)&((rpc_if_rep_t *)srvsvc_v3_0_s_ifspec)->id)) {
		return srvsvc_v3_0_s_ifspec;
	} else if (!id_compare(p_interface_id, (char *)&((rpc_if_rep_t *)dssetup_v0_0_s_ifspec)->id)) {
		return dssetup_v0_0_s_ifspec;
//	} else if (!id_compare(p_interface_id, (char *)&((rpc_if_rep_t *)lsa_v0_0_s_ifspec)->id)) {
//		return lsa_v0_0_s_ifspec;
	} else if (!id_compare(p_interface_id, (char *)&((rpc_if_rep_t *)wkssvc_v1_0_s_ifspec)->id)) {
		return wkssvc_v1_0_s_ifspec;
	}

	printf("ERROR: Unknown endpoint %s interface id %s\n", endpoint, p_interface_id);
	return NULL;
}

void hvrpc_capture_result(rpc_call_handle_t call_h, rpc_iovector_p_t response, unsigned32 *p_error) {
	int i=0, err = 0;
	hvrpc_call_t *hvhandle_p = (hvrpc_call_t *)call_h;
	rpc_iovector_elt_p_t elt0 = &response->elt[0];
	char *outbuf = NULL, *ptr = NULL;
	uint32_t outsize = 0;

	printf("%s call %p return data elements %d %p length %d error %d\n", __FUNCTION__, call_h, response->num_elt, elt0->data_addr, elt0->data_len, *p_error);
	/* Calculate the result buffer size */
	for (i=0; i< response->num_elt; i++) {
		outsize += response->elt[i].data_len;
	}
	printf("%s numelt %d length %d \n", __FUNCTION__, response->num_elt, outsize);

	/* Add space for the cnpkt header */
	outsize += sizeof(hvrpc_common_hdr_t);
	outbuf = hvbuf_malloc(outsize);
	if (!outbuf) {
		err = -1;
		goto cleanup;
	}

	/* Copy the result buffer from iovec to buffer */
	ptr = outbuf + sizeof(hvrpc_common_hdr_t);
	for (i=0; i< response->num_elt; i++) {
		memcpy(ptr, response->elt[i].data_addr, response->elt[i].data_len);
		ptr += response->elt[i].data_len;
	}

cleanup:
	if (err != 0) {
		if (outbuf) {
			free(outbuf);
			outbuf = NULL;
			outsize = 0;
		}
	}

	if (hvhandle_p) {
		hvhandle_p->error = err;
		hvhandle_p->response_buffer = outbuf;
		hvhandle_p->response_length = outsize;
	}
}

void hvrpc_flipendian_pkt(hvrpc_common_hdr_t *cnpkt) {
	SWAP16(cnpkt->frag_len);
	SWAP16(cnpkt->auth_len);
	SWAP32(cnpkt->call_id);
	SWAP32(cnpkt->alloc_hint);
	SWAP16(cnpkt->context_id);
	SWAP16(cnpkt->opnum);
}

hvrpc_common_hdr_t *hvrpc_parse_pkt(hvrpc_call_t *call_handle, unsigned8 *buffer, unsigned32 length, rpc_iovector_elt_p_t rpc_payload) {
	if (rpc_payload) {
		memset(rpc_payload, 0, sizeof(*rpc_payload));
	}

	if (!call_handle) {
		// Invalid handle
		return NULL;
	}

	memset(call_handle, 0, sizeof (*call_handle));
	if (length < sizeof(hvrpc_common_hdr_t)) {
		// Packet smaller than header
		return NULL;
	}

	hvrpc_common_hdr_t *cnpkt = (hvrpc_common_hdr_t *) buffer;
	call_handle->IDL_drep = ndr_g_local_drep;
	if (NDR_DREP_INT_REP(cnpkt->drep) !=  ndr_g_local_drep.int_rep) {
		call_handle->IDL_drep.int_rep = !call_handle->IDL_drep.int_rep;
		hvrpc_flipendian_pkt(cnpkt);
	}

	if (rpc_payload) {
		rpc_payload->data_addr = (unsigned8 *)&cnpkt[1];
		rpc_payload->data_len  = length - sizeof(cnpkt[0]);
	}
	return cnpkt;
}

int hvrpc_handle_pkt(hvrpc_iface_handle_t _ifspec, uint8_t *buffer, uint32_t length, char **p_outbuf, uint32_t *p_outsize) {
/*
 * TODO: Handle errors with hvnas status
 */
	char handle[1024];
	hvrpc_call_t hvhandle;
	ndr_format_t IDL_drep;
	rpc_if_rep_p_t ifspec = (rpc_if_rep_p_t)_ifspec;
	rpc_v2_server_stub_proc_t stub = NULL;
	rpc_call_handle_t myhandle = (rpc_call_handle_t)&hvhandle;
	error_status_t status = 0;
	rpc_iovector_elt_t iovec;

	memset(handle, 0, sizeof(handle));
	if (!p_outbuf || !p_outsize) {
		return -1;
	}

	hvrpc_common_hdr_t *cnpkt = hvrpc_parse_pkt(&hvhandle, (unsigned8 *)buffer, length, &iovec);
	if (!cnpkt) {
		return -2;
	}


	IDL_drep = hvhandle.IDL_drep;

	// TODO: Check opnum is in range
	stub = ifspec->server_epv[cnpkt->opnum];

	printf("server ifspec %p drep %d opnum %d stub %p\n", ifspec, IDL_drep.int_rep, cnpkt->opnum, stub);

	stub((handle_t)handle, myhandle, &iovec, &IDL_drep, NULL, ifspec->manager_epv, &status);
	if (hvhandle.error != 0) {
		return -3;
	}
	if (!hvhandle.response_buffer || !hvhandle.response_length) {
		return -4;
	}

	/* Create the CNPKT header for the response */
	hvrpc_common_hdr_t *opkt = (hvrpc_common_hdr_t *)hvhandle.response_buffer;
	memcpy(opkt->drep, cnpkt->drep, sizeof(opkt->drep));
	opkt->rpc_vers = cnpkt->rpc_vers;
	opkt->rpc_vers_minor = cnpkt->rpc_vers_minor;
	opkt->auth_len = 0;
	opkt->frag_len = 0xffffffff & hvhandle.response_length;
	opkt->call_id = cnpkt->call_id;
	opkt->ptype = DCERPC_PKT_RESPONSE;
	opkt->flags = 0;
	opkt->flags |= DCERPC_PFC_FLAG_FIRST;
	if (opkt->frag_len == hvhandle.response_length) {
		opkt->flags |= DCERPC_PFC_FLAG_LAST;
	}
	opkt->alloc_hint = opkt->frag_len - sizeof(hvrpc_common_hdr_t);
	opkt->context_id = cnpkt->context_id;
	opkt->opnum = cnpkt->opnum;

	if (hvhandle.IDL_drep.int_rep != ndr_g_local_drep.int_rep) {
		hvrpc_flipendian_pkt(opkt);
	}

	/* Return the result buffer */
	*p_outbuf = hvhandle.response_buffer;
	*p_outsize = hvhandle.response_length;

	return 0;
}


void hvswap16(unsigned8 *val16_p) {
	SWAP(val16_p[0], val16_p[1]);
}

void hvswap32(unsigned8 *val32_p) {
	SWAP(val32_p[0], val32_p[3]);
	SWAP(val32_p[1], val32_p[2]);
}

wchar16_t *hvrpc_string(const char *string) {
	int i=0;
	int nbytes = sizeof(wchar16_t) * (strlen(string) + 1);
	wchar16_t *result = malloc(nbytes);
	if (result) {
		memset(result, 0, nbytes);
		for (i=0; i<strlen(string); i++) {
			result[i] = string[i];
		}
	}

	return result;
}

int id_compare(const char *id1, const char *id2) {
	return memcmp(id1, id2, 16);
}
